@@ -11,7 +11,8 @@
 
 """
 @package gribapi
-@brief This package is the \b Python 3 interface to ecCodes. It offers almost one to one bindings to the C API functions.
+@brief This package is the \b Python 3 interface to ecCodes.
+It offers almost one to one bindings to the C API functions.
 
 The Python 3 interface to ecCodes uses the <a href="http://numpy.scipy.org/"><b>NumPy</b></a> package
 as the container of choice for the possible arrays of values that can be encoded/decoded in and from a message.
@@ -24,14 +25,16 @@ Numpy is a package used for scientific computing in Python and an efficient cont
 
 """
 
-from functools import wraps
-import sys
 import os
+import sys
+from functools import wraps
+
 import numpy as np
 
-from .bindings import ENC, ffi, lib
-from .bindings import __version__ as bindings_version  # noqa
 from . import errors
+from .bindings import ENC
+from .bindings import __version__ as bindings_version  # noqa
+from .bindings import ffi, lib, library_path
 
 try:
     type(file)
@@ -64,6 +67,10 @@ CODES_PRODUCT_TAF = 5
 GRIB_MISSING_DOUBLE = -1e100
 GRIB_MISSING_LONG = 2147483647
 
+# Constants for GRIB nearest neighbour
+GRIB_NEAREST_SAME_GRID = 1 << 0
+GRIB_NEAREST_SAME_DATA = 1 << 1
+GRIB_NEAREST_SAME_POINT = 1 << 2
 
 # ECC-1029: Disable function-arguments type-checking unless
 # environment variable is defined and equal to 1
@@ -237,7 +244,7 @@ def gts_new_from_file(fileobj, headers_only=False):
     @return               id of the GTS loaded in memory or None
     @exception CodesInternalError
     """
-    # err, h = err_last(lib.gts_new_from_file)(ffi.NULL, fileobj)
+
     err, h = err_last(lib.codes_handle_new_from_file)(
         ffi.NULL, fileobj, CODES_PRODUCT_GTS
     )
@@ -266,7 +273,7 @@ def metar_new_from_file(fileobj, headers_only=False):
     @return               id of the METAR loaded in memory or None
     @exception CodesInternalError
     """
-    # err, h = err_last(lib.metar_new_from_file)(ffi.NULL, fileobj)
+
     err, h = err_last(lib.codes_handle_new_from_file)(
         ffi.NULL, fileobj, CODES_PRODUCT_METAR
     )
@@ -393,7 +400,7 @@ def grib_new_from_file(fileobj, headers_only=False):
     @return               id of the grib loaded in memory or None
     @exception CodesInternalError
     """
-    # err, h = err_last(lib.grib_new_from_file)(ffi.NULL, fileobj, headers_only)
+
     err, h = err_last(lib.codes_handle_new_from_file)(
         ffi.NULL, fileobj, CODES_PRODUCT_GRIB
     )
@@ -484,7 +491,7 @@ def grib_get_string(msgid, key):
     length_p = ffi.new("size_t *", length)
     err = lib.grib_get_string(h, key.encode(ENC), values, length_p)
     GRIB_CHECK(err)
-    return ffi.string(values, length_p[0]).decode(ENC)
+    return _decode_bytes(values, length_p[0])
 
 
 @require(msgid=int, key=str, value=str)
@@ -813,7 +820,6 @@ def grib_keys_iterator_delete(iterid):
     @param iterid      keys iterator id created with @ref grib_keys_iterator_new
     @exception CodesInternalError
     """
-    # aa: THIS LEAKS MEMORY as it doesn't free all the connected iterators
     kih = get_grib_keys_iterator(iterid)
     lib.grib_keys_iterator_delete(kih)
 
@@ -829,7 +835,6 @@ def grib_keys_iterator_get_name(iterid):
     @return          key name to be retrieved
     @exception CodesInternalError
     """
-    # aa: missing call to grib_keys_iterator_get_accessor
     kih = get_grib_keys_iterator(iterid)
     name = lib.grib_keys_iterator_get_name(kih)
     return ffi.string(name).decode(ENC)
@@ -1169,6 +1174,19 @@ def grib_get_double_array(msgid, key):
     return arr
 
 
+# See ECC-1246
+def _decode_bytes(binput, maxlen=None):
+    if maxlen:
+        a_str = ffi.string(binput, maxlen)
+    else:
+        a_str = ffi.string(binput)
+    # Check for a MISSING value i.e., each character has all its bits=1
+    if all(x == 255 for x in a_str):
+        return ""
+    # Replace with a suitable replacement character rather than throw an exception
+    return a_str.decode(ENC, "replace")
+
+
 @require(msgid=int, key=str)
 def grib_get_string_array(msgid, key):
     """
@@ -1187,7 +1205,7 @@ def grib_get_string_array(msgid, key):
     size_p = ffi.new("size_t *", size)
     err = lib.grib_get_string_array(h, key.encode(ENC), values, size_p)
     GRIB_CHECK(err)
-    return [ffi.string(values[i]).decode(ENC) for i in range(size_p[0])]
+    return [_decode_bytes(values[i]) for i in range(size_p[0])]
 
 
 @require(msgid=int, key=str)
@@ -1313,7 +1331,10 @@ def grib_index_new_from_file(filename, keys):
 
     @param filename   path of the file to index on
     @param keys       sequence of keys to index on.
-                      The type of the key can be explicitly declared appending :l for long (or alternatively :i), :d for double, :s for string to the key name.
+                      The type of the key can be explicitly declared appending
+                      :l for long (or alternatively :i),
+                      :d for double,
+                      :s for string to the key name.
     @return           index id
     @exception CodesInternalError
     """
@@ -1381,7 +1402,8 @@ def grib_index_get_long(indexid, key):
     @brief Get the distinct values of the key in argument contained in the index.
     The key must belong to the index.
 
-    This function is used when the type of the key was explicitly defined as long or when the native type of the key is long.
+    This function is used when the type of the key was explicitly defined as long or when the native type of
+    the key is long.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1406,7 +1428,8 @@ def grib_index_get_string(indexid, key):
     @brief Get the distinct values of the key in argument contained in the index.
     The key must belong to the index.
 
-    This function is used when the type of the key was explicitly defined as string or when the native type of the key is string.
+    This function is used when the type of the key was explicitly defined as string or when the native type of
+    the key is string.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1432,7 +1455,8 @@ def grib_index_get_double(indexid, key):
     @brief Get the distinct values of the key in argument contained in the index.
     The key must belong to the index.
 
-    This function is used when the type of the key was explicitly defined as double or when the native type of the key is double.
+    This function is used when the type of the key was explicitly defined as double or when the native type
+    of the key is double.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1446,9 +1470,9 @@ def grib_index_get_double(indexid, key):
 
     values_p = ffi.new("double[]", nval)
     size_p = ffi.new("size_t *", nval)
-    err = lib.grib_index_get_doule(ih, key.encode(ENC), values_p, size_p)
+    err = lib.grib_index_get_double(ih, key.encode(ENC), values_p, size_p)
     GRIB_CHECK(err)
-    return tuple(int(values_p[i]) for i in range(size_p[0]))
+    return tuple(values_p[i] for i in range(size_p[0]))
 
 
 @require(indexid=int, key=str, value=int)
@@ -1457,7 +1481,8 @@ def grib_index_select_long(indexid, key, value):
     @brief Select the message subset with key==value.
     The value is an integer.
 
-    The key must have been created with integer type or have integer as native type if the type was not explicitly defined in the index creation.
+    The key must have been created with integer type or have integer as native type if the type
+    was not explicitly defined in the index creation.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1476,7 +1501,8 @@ def grib_index_select_double(indexid, key, value):
     @brief Select the message subset with key==value.
     The value is a double.
 
-    The key must have been created with integer type or have integer as native type if the type was not explicitly defined in the index creation.
+    The key must have been created with integer type or have integer as native type if the type was
+    not explicitly defined in the index creation.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1495,7 +1521,8 @@ def grib_index_select_string(indexid, key, value):
     @brief Select the message subset with key==value.
     The value is an integer.
 
-    The key must have been created with string type or have string as native type if the type was not explicitly defined in the index creation.
+    The key must have been created with string type or have string as native type if the type
+    was not explicitly defined in the index creation.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1514,7 +1541,8 @@ def grib_new_from_index(indexid):
     @brief Create a new handle from an index after having selected the key values.
 
     All the keys belonging to the index must be selected before calling this function.
-    Successive calls to this function will return all the handles compatible with the constraints defined selecting the values of the index keys.
+    Successive calls to this function will return all the handles compatible with the constraints
+    defined selecting the values of the index keys.
 
     The message can be accessed through its gribid and will be available until @ref grib_release is called.
 
@@ -1582,7 +1610,7 @@ def grib_get_double_element(msgid, key, index):
     """
     h = get_handle(msgid)
     value_p = ffi.new("double*")
-    err = lib.grib_get_double_element(h, key.encode(ENC), index, value_p)  # TODO
+    err = lib.grib_get_double_element(h, key.encode(ENC), index, value_p)
     GRIB_CHECK(err)
     return value_p[0]
 
@@ -1816,7 +1844,8 @@ def grib_find_nearest(gribid, inlat, inlon, is_lsm=False, npoints=1):
 @require(gribid=int, is_lsm=bool)
 def grib_find_nearest_multiple(gribid, is_lsm, inlats, inlons):
     """
-    @brief Find the nearest point of a set of points whose latitudes and longitudes are given in the inlats, inlons arrays respectively
+    @brief Find the nearest point of a set of points whose latitudes and longitudes are given in
+    the inlats, inlons arrays respectively
 
     @param gribid     id of the GRIB message loaded in memory
     @param is_lsm     True if the nearest land point is required otherwise False.
@@ -2041,8 +2070,11 @@ def grib_set(msgid, key, value):
     #    # The value passed in is iterable; i.e. a list or array etc
     #    grib_set_array(msgid, key, value)
     else:
+        hint = ""
+        if hasattr(value, "__iter__"):
+            hint = " (Hint: for array keys use codes_set_array(msgid, key, value))"
         raise errors.GribInternalError(
-            "Invalid type of value when setting key '%s'." % key
+            "Invalid type of value when setting key '%s'%s." % (key, hint)
         )
 
 
@@ -2319,6 +2351,13 @@ def grib_set_samples_path(samples_path):
     lib.grib_context_set_samples_path(context, samples_path.encode(ENC))
 
 
+def grib_context_delete():
+    """
+    @brief Wipe all the cached data and definitions files in the context
+    """
+    lib.grib_context_delete(ffi.NULL)
+
+
 def codes_bufr_multi_element_constant_arrays_on():
     """
     @brief BUFR: Turn on the mode where you get multiple elements
@@ -2341,6 +2380,19 @@ def codes_bufr_multi_element_constant_arrays_off():
     lib.codes_bufr_multi_element_constant_arrays_off(context)
 
 
+@require(msgid=int)
+def codes_dump(msgid, output_fileobj=sys.stdout, mode="wmo", flags=0):
+    """
+    @brief Print all keys to an output file object, with the given dump mode and flags
+
+    @param msgid          id of the message loaded in memory
+    @param output_fileobj output file object e.g., sys.stdout
+    @param mode           dump mode e.g., "wmo", "debug", "json"
+    """
+    h = get_handle(msgid)
+    lib.grib_dump_content(h, output_fileobj, mode.encode(ENC), flags, ffi.NULL)
+
+
 # Convert the C codes_bufr_header struct to a Python dictionary
 def _convert_struct_to_dict(s):
     result = {}
@@ -2356,7 +2408,7 @@ def _convert_struct_to_dict(s):
 
 def codes_bufr_extract_headers(filepath, is_strict=True):
     """
-    @brief BUFR header extraction (EXPERIMENTAL FEATURE)
+    @brief BUFR header extraction
 
     @param filepath       path of input BUFR file
     @param is_strict      fail as soon as any invalid BUFR message is encountered
@@ -2387,6 +2439,7 @@ def codes_bufr_extract_headers(filepath, is_strict=True):
         i += 1
 
 
+@require(msgid=int)
 def codes_bufr_key_is_header(msgid, key):
     """
     @brief Check if the BUFR key is in the header or in the data section.
@@ -2407,11 +2460,12 @@ def codes_bufr_key_is_header(msgid, key):
 
 def codes_extract_offsets(filepath, product_kind, is_strict=True):
     """
-    @brief Message offset extraction (EXPERIMENTAL FEATURE)
+    @brief Message offset extraction
 
-    @param filepath       path of input file
-    @param is_strict      fail as soon as any invalid message is encountered
-    @return               a list of offsets
+    @param filepath   path of input file
+    @product_kind     one of CODES_PRODUCT_GRIB, CODES_PRODUCT_BUFR, CODES_PRODUCT_ANY or CODES_PRODUCT_GTS
+    @param is_strict  if True, fail as soon as any invalid message is encountered
+    @return           a generator that yields offsets (each offset is an integer)
     @exception CodesInternalError
     """
     context = lib.grib_context_get_default()
@@ -2430,3 +2484,84 @@ def codes_extract_offsets(filepath, product_kind, is_strict=True):
     while i < num_messages:
         yield offsets[i]
         i += 1
+
+
+# -------------------------------
+# EXPERIMENTAL FEATURES
+# -------------------------------
+@require(msgid=int)
+def grib_nearest_new(msgid):
+    h = get_handle(msgid)
+    err, nid = err_last(lib.grib_nearest_new)(h)
+    GRIB_CHECK(err)
+    return put_grib_nearest(nid)
+
+
+def put_grib_nearest(nid):
+    return int(ffi.cast("size_t", nid))
+
+
+def get_grib_nearest(nid):
+    return ffi.cast("grib_nearest*", nid)
+
+
+@require(nid=int)
+def grib_nearest_delete(nid):
+    nh = get_grib_nearest(nid)
+    lib.grib_nearest_delete(nh)
+
+
+@require(nid=int, gribid=int)
+def grib_nearest_find(nid, gribid, inlat, inlon, flags, is_lsm=False, npoints=4):
+    # flags has to be one of:
+    #  GRIB_NEAREST_SAME_GRID
+    #  GRIB_NEAREST_SAME_DATA
+    #  GRIB_NEAREST_SAME_POINT
+    if npoints != 4:
+        raise errors.FunctionNotImplementedError(
+            "grib_nearest_find npoints argument: Only 4 points supported"
+        )
+    if is_lsm:
+        raise errors.FunctionNotImplementedError(
+            "grib_nearest_find is_lsm argument: Land sea mask not supported"
+        )
+
+    h = get_handle(gribid)
+    outlats_p = ffi.new("double[]", npoints)
+    outlons_p = ffi.new("double[]", npoints)
+    values_p = ffi.new("double[]", npoints)
+    distances_p = ffi.new("double[]", npoints)
+    indexes_p = ffi.new("int[]", npoints)
+    size = ffi.new("size_t *")
+    nh = get_grib_nearest(nid)
+    err = lib.grib_nearest_find(
+        nh,
+        h,
+        inlat,
+        inlon,
+        flags,
+        outlats_p,
+        outlons_p,
+        values_p,
+        distances_p,
+        indexes_p,
+        size,
+    )
+    GRIB_CHECK(err)
+    result = []
+    for i in range(npoints):
+        result.append(
+            Bunch(
+                lat=outlats_p[i],
+                lon=outlons_p[i],
+                value=values_p[i],
+                distance=distances_p[i],
+                index=indexes_p[i],
+            )
+        )
+
+    return tuple(result)
+
+
+def codes_get_library_path():
+    return library_path
